#define __CL_ENABLE_EXCEPTIONS

#include <fstream>
#include <iostream>
#include <iterator>
#include <CL/cl.hpp>
#include <CL/opencl.h>

using namespace std;
int main () {

    vector<cl::Platform> platforms;
    vector<cl::Device> devices;
    vector<cl::Kernel> kernels;
    
    try {
    
        // create platform
        cl::Platform::get(&platforms);
        platforms[0].getDevices(CL_DEVICE_TYPE_GPU, &devices);

        // create context
        cl::Context context(devices);

        // create command queue
        cl::CommandQueue queue(context, devices[0]);

        // load opencl source
        ifstream cl_file("pbkdf2_hmac_sha1_unsplit_kernel.cl");
        string cl_string(istreambuf_iterator<char>(cl_file), (istreambuf_iterator<char>()));
        cl::Program::Sources source(1, make_pair(cl_string.c_str(), 
            cl_string.length() + 1));

        // create program
        cl::Program program(context, source);

        // compile opencl source
        program.build(devices);

        // load named kernel from opencl source
        cl::Kernel kernel(program, "PBKDF2");

        // create a message to send to kernel ----------------------------------------------------------------
        char* password = "C & OpenCL is cool";
        //unsigned int keyBytes = 18; // keyBytes
        cl_uint keyBytes = 18; // keyBytes  

        char* passwordSalt = "12345678901234567890123456789012";
        cl_uint passwordSaltLen = 32;

//        cl_int passwordIterations = 2;
        cl_uint passwordIterations = 2;

        cl_uint bufferC_size = 20;

        // ---------------------------------------------------------------------------------------------------
        // allocate device buffer to hold message
        cl::Buffer bufferA(context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, sizeof(char) * keyBytes, password);
        cl::Buffer bufferB(context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, sizeof(char) * passwordSaltLen, passwordSalt);
        cl::Buffer bufferC(context, CL_MEM_WRITE_ONLY, sizeof(char) * keyBytes);

        queue.enqueueWriteBuffer(bufferA, CL_TRUE, 0, keyBytes * sizeof(char), password);
        queue.enqueueWriteBuffer(bufferB, CL_TRUE, 0, passwordSaltLen * sizeof(char), passwordSalt);

/*
inline void pbkdf2(__global const uchar * pass, uint passlen,
                   __global const uchar * salt, uint saltlen, uint iterations,
                   __global uint * out, uint outlen)
*/
        // set message as kernel argument
        kernel.setArg(0, bufferA);
        kernel.setArg(1, keyBytes);
        kernel.setArg(2, bufferB);
        kernel.setArg(3, passwordSaltLen); // saltlen
        kernel.setArg(4, passwordIterations);
        kernel.setArg(5, bufferC);
        kernel.setArg(6, bufferC_size);

         // execute kernel
        queue.enqueueTask(kernel);

        // wait for completion
        queue.finish();
        // ----------------------
        char* out_global = "                 !";
 
	queue.enqueueReadBuffer(bufferC, CL_TRUE, 0, keyBytes * sizeof(char), out_global);
        cout << "Output " << out_global << endl;
        // ----------------------
        cout << endl;
        
    } catch (cl::Error e) {
        cout << endl << e.what() << " : " << e.err() << endl;
    }
    
    return 0;
    
}

